#ifndef MATHLIB_MATRIX_TPP
#define MATHLIB_MATRIX_TPP

// implementation of matrix and vector

#include <mathlib/matrix.h>
#include <mathlib/mathops.tpp>
#include <mathlib/utility.h>

#if MATHLIB_NO_OSTREAM
#	include <iostream>
#else
#	include <ostream>
#endif

namespace mathlib {

typedef basic_matrix<2, 1> bm21;
typedef basic_matrix<3, 1> bm31;
typedef basic_matrix<4, 1> bm41;
typedef basic_matrix<1, 2> bm12;
typedef basic_matrix<2, 2> bm22;
typedef basic_matrix<3, 2> bm32;
typedef basic_matrix<4, 2> bm42;
typedef basic_matrix<1, 3> bm13;
typedef basic_matrix<2, 3> bm23;
typedef basic_matrix<3, 3> bm33;
typedef basic_matrix<4, 3> bm43;
typedef basic_matrix<1, 4> bm14;
typedef basic_matrix<2, 4> bm24;
typedef basic_matrix<3, 4> bm34;
typedef basic_matrix<4, 4> bm44;

template<class, class> class matrix_unop;
template<class, class, class> class matrix_binop;

// float constant expression
struct float_keeper {
	static const int rows = 1;
	static const int columns = 1;
	static const bool is_storage = true;
	static const bool is_lvalue = false;
	static const bool error = false;
	static const bool element_op = true;
	typedef float_keeper base_type;
	float f;
	float_keeper() {}
	float_keeper(float v) : f(v) {}
	template<int, int> float element() const { return f; }
};

// used for assignment
struct zero_assigner_type {
	typedef zero_assigner_type base_type;
	template<int, int> static float element() { return 0.f; }
};
struct identity_assigner_type {
	typedef identity_assigner_type base_type;
	template<int R, int C> static float element() { return R == C; }
};

// base class for axis constant types
struct axis_type {
	static const int rows = 3;
	static const int columns = 1;
	static const bool is_storage = true;
	static const bool is_lvalue = false;
	static const bool error = false;
	static const bool element_op = true;
};

// axis constant types
struct x_axis_type : axis_type {
	typedef x_axis_type base_type;
	template<int R, int C> static float element() { return R == 0; }
};
struct y_axis_type : axis_type {
	typedef y_axis_type base_type;
	template<int R, int C> static float element() { return R == 1; }
};
struct z_axis_type : axis_type {
	typedef z_axis_type base_type;
	template<int R, int C> static float element() { return R == 2; }
};

// fake axis vector constants
const tmatrix<x_axis_type> x_axis = tmatrix<x_axis_type>();
const tmatrix<y_axis_type> y_axis = tmatrix<y_axis_type>();
const tmatrix<z_axis_type> z_axis = tmatrix<z_axis_type>();

// find out which type should be used to hold references to expressions
// (const basic_matrix &, all others as themselves)
template<class T> struct matrix_ref {
	typedef const T type;
	typedef T noconst_type;
};
template<int A, int B> struct matrix_ref<basic_matrix<A, B> > {
	typedef const basic_matrix<A, B> & MATHLIB_RESTRICT type;
	typedef basic_matrix<A, B> & MATHLIB_RESTRICT noconst_type;
};

template<class T, class U, bool Is_scalar>
struct select_matrix_helper { typedef T type; };
template<class T, class U>
struct select_matrix_helper<T, U, true> { typedef U type; };

// select the leftmost argument that is a matrix
template<class T, class U>
struct select_matrix {
	enum { ok = T::rows == 1 && T::columns == 1 };
	typedef typename select_matrix_helper<T, U, ok>::type type;
};

// for convenience. check if argument is a 3x1 matrix
template<class T> struct is_vector_3
	{ enum { value = T::rows == 3 && T::columns == 1 }; };

// the type of temporary storage required.
template<class T, bool Is_scalar>
struct result_type_helper {
	typedef basic_matrix<T::rows, T::columns> type;
};
template<class T> struct result_type_helper<T, true> {
	typedef float_keeper type;
};

template<class T> struct result_type
	: result_type_helper<T, (T::rows == 1 && T::columns == 1)> {};

// check size equality
template<int R1, int C1, int R2, int C2> struct assert_size {};
template<int R, int C> struct assert_size<R, C, R, C> { enum { ok = 1 }; };

} // namespace mathlib

#include <mathlib/matrix_indexing.tpp>
#include <mathlib/matrix_unop.tpp>
#include <mathlib/matrix_binop.tpp>
#include <mathlib/simd/implementation.tpp>
#include <mathlib/matrix_operators.tpp>
#include <mathlib/matrix_assign.tpp>
#include <mathlib/matrix_inline.tpp>

#endif
